@@ -4,9 +4,9 @@
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
 import contextlib
-from distutils.version import LooseVersion
+from packaging.version import Version
 import errno
-import glob
+from glob import glob
 import hashlib
 import itertools
 import json
@@ -25,18 +25,19 @@ from urllib.request import Request, urlopen, urljoin, quote
 from urllib.error import URLError
 import warnings
 
-from jupyter_core.paths import jupyter_config_path
+from jupyter_core.paths import jupyter_config_path, jupyter_path
 from jupyterlab_server.process import which, Process, WatchHelper, list2cmdline
-from notebook.nbextensions import GREEN_ENABLED, GREEN_OK, RED_DISABLED, RED_X
-from traitlets import HasTraits, Bool, Unicode, Instance, default
+from jupyterlab_server.config import LabConfig, get_page_config, get_federated_extensions, get_static_page_config, write_page_config
+from jupyter_server.extension.serverextension import GREEN_ENABLED, GREEN_OK, RED_DISABLED, RED_X
+from traitlets import HasTraits, Bool, Dict, Instance, List, Unicode, default
 
-from .semver import Range, gte, lt, lte, gt, make_semver
-from .jlpmapp import YARN_PATH, HERE
-from .coreconfig import _get_default_core_data, CoreConfig
+from jupyterlab.semver import Range, gte, lt, lte, gt, make_semver
+from jupyterlab.jlpmapp import YARN_PATH, HERE
+from jupyterlab.coreconfig import _get_default_core_data, CoreConfig
 
 
 # The regex for expecting the webpack output.
-WEBPACK_EXPECT = re.compile(r'.*/index.out.js')
+WEBPACK_EXPECT = re.compile(r'.*theme-light-extension/style/index.css')
 
 # The dev mode directory.
 DEV_DIR = osp.abspath(os.path.join(HERE, '..', 'dev_mode'))
@@ -323,13 +324,15 @@ class AppOptions(HasTraits):
 
     kill_event = Instance(Event, args=(), help='Event for aborting call')
 
+    labextensions_path = List(Unicode(), help='The paths to look in for prebuilt JupyterLab extensions')
+
     registry = Unicode(help="NPM packages registry URL")
 
     @default('logger')
     def _default_logger(self):
         return logging.getLogger('jupyterlab')
 
-    # These defaults need to be dynamic to pick up
+    # These defaults need to be federated to pick up
     # any changes to env vars:
     @default('app_dir')
     def _default_app_dir(self):
@@ -450,40 +453,41 @@ def clean(app_options=None):
 
 
 def build(name=None, version=None, static_url=None,
-          command='build:prod', kill_event=None,
-          clean_staging=False, app_options=None):
+          kill_event=None,
+          clean_staging=False, app_options=None, production=True, minimize=True):
     """Build the JupyterLab application.
     """
     app_options = _ensure_options(app_options)
     _node_check(app_options.logger)
     handler = _AppHandler(app_options)
     return handler.build(name=name, version=version, static_url=static_url,
-                         command=command, clean_staging=clean_staging)
+                         production=production, minimize=minimize, clean_staging=clean_staging)
 
 
 def get_app_info(app_options=None):
     """Get a dictionary of information about the app.
     """
     handler = _AppHandler(app_options)
+    handler._ensure_disabled_info()
     return handler.info
 
 
-def enable_extension(extension, app_options=None):
+def enable_extension(extension, app_options=None, level='sys_prefix'):
     """Enable a JupyterLab extension.
 
     Returns `True` if a rebuild is recommended, `False` otherwise.
     """
     handler = _AppHandler(app_options)
-    return handler.toggle_extension(extension, False)
+    return handler.toggle_extension(extension, False, level=level)
 
 
-def disable_extension(extension, app_options=None):
+def disable_extension(extension, app_options=None, level='sys_prefix'):
     """Disable a JupyterLab package.
 
     Returns `True` if a rebuild is recommended, `False` otherwise.
     """
     handler = _AppHandler(app_options)
-    return handler.toggle_extension(extension, True)
+    return handler.toggle_extension(extension, True, level=level)
 
 
 def check_extension(extension, installed=False, app_options=None):
@@ -566,14 +570,20 @@ class _AppHandler(object):
         """Create a new _AppHandler object
         """
         options = _ensure_options(options)
+        self._options = options
         self.app_dir = options.app_dir
         self.sys_dir = get_app_dir() if options.use_sys_dir else self.app_dir
         self.logger = options.logger
         self.core_data = options.core_config._data
-        self.info = self._get_app_info()
+        self.labextensions_path = options.labextensions_path
         self.kill_event = options.kill_event
         self.registry = options.registry
 
+        # Do this last since it relies on other attributes
+        self.info = self._get_app_info()
+
+
+
     def install_extension(self, extension, existing=None, pin=None):
         """Install an extension package into JupyterLab.
 
@@ -621,19 +631,19 @@ class _AppHandler(object):
         return True
 
     def build(self, name=None, version=None, static_url=None,
-              command='build:prod:minimize', clean_staging=False):
+              clean_staging=False, production=True, minimize=True):
         """Build the application.
         """
-        # resolve the build type
-        parts = command.split(':')
-        if len(parts) < 2:
-            parts.append('dev')
-        elif parts[1] == 'none':
-            parts[1] = ('dev' if self.info['linked_packages'] or self.info['local_extensions'] else
-                        'prod')
-        command = ':'.join(parts)
+        if production is None:
+            production = not (self.info['linked_packages'] or self.info['local_extensions'])
 
-        self.logger.info('Building jupyterlab assets (%s)' % command)
+        if not production:
+            minimize = False
+
+        info = ['production' if production else 'development']
+        if production:
+            info.append('minimized' if minimize else 'not minimized')
+        self.logger.info(f'Building jupyterlab assets ({", ".join(info)})')
 
         # Set up the build directory.
         app_dir = self.app_dir
@@ -652,10 +662,13 @@ class _AppHandler(object):
             self.logger.debug(msg)
             raise RuntimeError(msg)
 
-        dedupe_yarn(staging, self.logger)
-
         # Build the app.
-        ret = self._run(['node', YARN_PATH, 'run', command], cwd=staging)
+        dedupe_yarn(staging, self.logger)
+        starting_dir = os.getcwd()
+        os.chdir(staging)
+        command = f'build:{"prod" if production else "dev"}{":minimize" if minimize else ""}'
+        ret = self._run(['node', YARN_PATH, 'run', command])
+        os.chdir(starting_dir)
         if ret != 0:
             msg = 'JupyterLab failed to build'
             self.logger.debug(msg)
@@ -682,46 +695,68 @@ class _AppHandler(object):
     def list_extensions(self):
         """Print an output of the extensions.
         """
+        self._ensure_disabled_info()
         logger = self.logger
         info = self.info
 
-        print('JupyterLab v%s' % info['version'])
+        logger.info('JupyterLab v%s' % info['version'])
+
+        if info['federated_extensions'] or info['extensions']:
+            info['compat_errors'] = self._get_extension_compat()
+
+        if info['federated_extensions']:
+            self._list_federated_extensions()
 
         if info['extensions']:
-            info['compat_errors'] = self._get_extension_compat()
-            print('Known labextensions:')
+            logger.info('Other labextensions (built into JupyterLab)')
             self._list_extensions(info, 'app')
             self._list_extensions(info, 'sys')
-        else:
-            print('No installed extensions')
 
         local = info['local_extensions']
         if local:
-            print('\n   local extensions:')
+            logger.info('\n   local extensions:')
             for name in sorted(local):
-                print('        %s: %s' % (name, local[name]))
+                logger.info('        %s: %s' % (name, local[name]))
 
         linked_packages = info['linked_packages']
         if linked_packages:
-            print('\n   linked packages:')
+            logger.info('\n   linked packages:')
             for key in sorted(linked_packages):
                 source = linked_packages[key]['source']
-                print('        %s: %s' % (key, source))
+                logger.info('        %s: %s' % (key, source))
 
         uninstalled_core = info['uninstalled_core']
         if uninstalled_core:
-            print('\nUninstalled core extensions:')
-            [print('    %s' % item) for item in sorted(uninstalled_core)]
+            logger.info('\nUninstalled core extensions:')
+            [logger.info('    %s' % item) for item in sorted(uninstalled_core)]
 
-        disabled_core = info['disabled_core']
-        if disabled_core:
-            print('\nDisabled core extensions:')
-            [print('    %s' % item) for item in sorted(disabled_core)]
+        all_exts = list(info['federated_extensions']) + list(info['extensions']) + list(info['core_extensions'])
+        # Ignore disabled extensions that are not installed
+        disabled = [i for i in info['disabled'] if i.partition(':')[0] in all_exts]
+        if disabled:
+            logger.info('\nDisabled extensions:')
+            for item in sorted(disabled):
+                # Show that all plugins will be disabled if the whole extension matches
+                if item in all_exts:
+                    item += ' (all plugins)'
+                logger.info('    %s' % item)
+
+        # Here check if modules are improperly shadowed
+        improper_shadowed = []
+        for ext_name in self.info['shadowed_exts']:
+           source_version = self.info['extensions'][ext_name]['version']
+           prebuilt_version = self.info['federated_extensions'][ext_name]['version']
+           if not gte(prebuilt_version, source_version, True):
+               improper_shadowed.append(ext_name)
+
+        if improper_shadowed:
+            logger.info('\nThe following source extensions are overshadowed by older prebuilt extensions:')
+            [logger.info('    %s' % name) for name in sorted(improper_shadowed)]
 
         messages = self.build_check(fast=True)
         if messages:
-            print('\nBuild recommended, please run `jupyter lab build`:')
-            [print('    %s' % item) for item in messages]
+            logger.info('\nBuild recommended, please run `jupyter lab build`:')
+            [logger.info('    %s' % item) for item in messages]
 
     def build_check(self, fast=False):
         """Determine whether JupyterLab should be built.
@@ -745,10 +780,12 @@ class _AppHandler(object):
         # Look for mismatched version.
         static_version = old_jlab.get('version', '')
         core_version = old_jlab['version']
-        if LooseVersion(static_version) != LooseVersion(core_version):
+        if Version(static_version) != Version(core_version):
             msg = 'Version mismatch: %s (built), %s (current)'
             return [msg % (static_version, core_version)]
 
+        shadowed_exts = self.info['shadowed_exts']
+
         # Look for mismatched extensions.
         new_package = self._get_package_template(silent=fast)
         new_jlab = new_package['jupyterlab']
@@ -757,11 +794,15 @@ class _AppHandler(object):
         for ext_type in ['extensions', 'mimeExtensions']:
             # Extensions that were added.
             for ext in new_jlab[ext_type]:
+                if ext in shadowed_exts:
+                    continue
                 if ext not in old_jlab[ext_type]:
                     messages.append('%s needs to be included in build' % ext)
 
             # Extensions that were removed.
             for ext in old_jlab[ext_type]:
+                if ext in shadowed_exts:
+                    continue
                 if ext not in new_jlab[ext_type]:
                     messages.append('%s needs to be removed from build' % ext)
 
@@ -778,7 +819,7 @@ class _AppHandler(object):
 
         # Look for updated local extensions.
         for (name, source) in local.items():
-            if fast:
+            if fast or name in shadowed_exts:
                 continue
             dname = pjoin(app_dir, 'extensions')
             if self._check_local(name, source, dname):
@@ -786,7 +827,7 @@ class _AppHandler(object):
 
         # Look for updated linked packages.
         for (name, item) in linked.items():
-            if fast:
+            if fast or name in shadowed_exts:
                 continue
             dname = pjoin(app_dir, 'staging', 'linked_packages')
             if self._check_local(name, item['source'], dname):
@@ -799,25 +840,35 @@ class _AppHandler(object):
 
         Returns `True` if a rebuild is recommended, `False` otherwise.
         """
+        info = self.info
+        logger = self.logger
+
+        if name in info['federated_extensions']:
+            if info['federated_extensions'][name].get('install', dict()).get('uninstallInstructions', None):
+                logger.error('JupyterLab cannot uninstall this extension. %s' % info['federated_extensions'][name]['install']['uninstallInstructions'])
+            else:
+                logger.error('JupyterLab cannot uninstall %s since it was installed outside of JupyterLab. Use the same method used to install this extension to uninstall this extension.' % name)
+            return False
+
         # Allow for uninstalled core extensions.
-        if name in self.info['core_extensions']:
+        if name in info['core_extensions']:
             config = self._read_build_config()
             uninstalled = config.get('uninstalled_core_extensions', [])
             if name not in uninstalled:
-                self.logger.info('Uninstalling core extension %s' % name)
+                logger.info('Uninstalling core extension %s' % name)
                 uninstalled.append(name)
                 config['uninstalled_core_extensions'] = uninstalled
                 self._write_build_config(config)
                 return True
             return False
 
-        local = self.info['local_extensions']
+        local = info['local_extensions']
 
-        for (extname, data) in self.info['extensions'].items():
+        for (extname, data) in info['extensions'].items():
             path = data['path']
             if extname == name:
                 msg = 'Uninstalling %s from %s' % (name, osp.dirname(path))
-                self.logger.info(msg)
+                logger.info(msg)
                 os.remove(path)
                 # Handle local extensions.
                 if extname in local:
@@ -827,7 +878,7 @@ class _AppHandler(object):
                     self._write_build_config(config)
                 return True
 
-        self.logger.warn('No labextension named "%s" installed' % name)
+        logger.warn('No labextension named "%s" installed' % name)
         return False
 
     def uninstall_all_extensions(self):
@@ -950,27 +1001,35 @@ class _AppHandler(object):
         self._write_build_config(config)
         return True
 
-    def toggle_extension(self, extension, value):
+    def toggle_extension(self, extension, value, level='sys_prefix'):
         """Enable or disable a lab extension.
 
         Returns `True` if a rebuild is recommended, `False` otherwise.
         """
-        config = self._read_page_config()
-        disabled = config.setdefault('disabledExtensions', [])
+        lab_config = LabConfig()
+        app_settings_dir = osp.join(self.app_dir, 'settings')
+
+        page_config = get_static_page_config(app_settings_dir=app_settings_dir, logger=self.logger, level=level)
+
+        disabled = page_config.get('disabledExtensions', {})
         did_something = False
-        if value and extension not in disabled:
-            disabled.append(extension)
+        is_disabled = disabled.get(extension, False)
+        if value and not is_disabled:
+            disabled[extension] = True
             did_something = True
-        elif not value and extension in disabled:
-            disabled.remove(extension)
+        elif not value and is_disabled:
+            disabled[extension] = False
             did_something = True
+
         if did_something:
-            self._write_page_config(config)
+            page_config['disabledExtensions'] = disabled
+            write_page_config(page_config, level=level)
         return did_something
 
     def check_extension(self, extension, check_installed_only=False):
         """Check if a lab extension is enabled or disabled
         """
+        self._ensure_disabled_info()
         info = self.info
 
         if extension in info["core_extensions"]:
@@ -1030,8 +1089,7 @@ class _AppHandler(object):
         info = dict()
         info['core_data'] = core_data = self.core_data
         info['extensions'] = extensions = self._get_extensions(core_data)
-        page_config = self._read_page_config()
-        info['disabled'] = page_config.get('disabledExtensions', [])
+
         info['local_extensions'] = self._get_local_extensions()
         info['linked_packages'] = self._get_linked_packages()
         info['app_extensions'] = app = []
@@ -1053,16 +1111,30 @@ class _AppHandler(object):
         info['sys_dir'] = self.sys_dir
         info['app_dir'] = self.app_dir
 
-        info['core_extensions'] = core_extensions = _get_core_extensions(
-            self.core_data)
+        info['core_extensions'] = _get_core_extensions(self.core_data)
+
+        info['federated_extensions'] = get_federated_extensions(self.labextensions_path)
+        info['shadowed_exts'] = [ext for ext in info['extensions'] if ext in info['federated_extensions']]
+        return info
+
+    def _ensure_disabled_info(self):
+        info = self.info
+        if 'disabled' in info:
+            return
+
+        labextensions_path = self.labextensions_path
+        app_settings_dir = osp.join(self.app_dir, 'settings')
+
+        page_config = get_page_config(labextensions_path, app_settings_dir=app_settings_dir, logger=self.logger)
+
+        info['disabled'] = page_config.get('disabledExtensions', [])
 
         disabled_core = []
-        for key in core_extensions:
+        for key in info['core_extensions']:
             if key in info['disabled']:
                 disabled_core.append(key)
 
         info['disabled_core'] = disabled_core
-        return info
 
     def _populate_staging(self, name=None, version=None, static_url=None,
                           clean=False):
@@ -1088,7 +1160,8 @@ class _AppHandler(object):
                 _rmtree(staging, self.logger)
                 os.makedirs(staging)
 
-        for fname in ['index.js', 'webpack.config.js',
+        for fname in ['index.js', 'bootstrap.js', 'publicpath.js',
+                      'webpack.config.js',
                       'webpack.prod.config.js',
                       'webpack.prod.minimize.config.js',
                       '.yarnrc', 'yarn.js']:
@@ -1181,6 +1254,7 @@ class _AppHandler(object):
         local = self.info['local_extensions']
         linked = self.info['linked_packages']
         extensions = self.info['extensions']
+        shadowed_exts = self.info['shadowed_exts']
         jlab = data['jupyterlab']
 
         def format_path(path):
@@ -1194,17 +1268,23 @@ class _AppHandler(object):
 
         # Handle local extensions.
         for (key, source) in local.items():
+            if key in shadowed_exts:
+                continue
             jlab['linkedPackages'][key] = source
-            data['resolutions'][key] = source
+            data['resolutions'][key] = 'file:' + self.info['extensions'][key]['path']
 
         # Handle linked packages.
         for (key, item) in linked.items():
+            if key in shadowed_exts:
+                continue
             path = pjoin(self.app_dir, 'staging', 'linked_packages')
             path = pjoin(path, item['filename'])
             data['dependencies'][key] = format_path(path)
             jlab['linkedPackages'][key] = item['source']
             data['resolutions'][key] = format_path(path)
 
+        data['jupyterlab']['extensionMetadata'] = dict()
+
         # Handle extensions
         compat_errors = self._get_extension_compat()
         for (key, value) in extensions.items():
@@ -1228,6 +1308,9 @@ class _AppHandler(object):
                     ext = ''
                 jlab[item + 's'][key] = ext
 
+                # Add metadata for the extension
+                data['jupyterlab']['extensionMetadata'][key] = jlab_data
+
         # Handle uninstalled core extensions.
         for item in self.info['uninstalled_core']:
             if item in jlab['extensions']:
@@ -1306,7 +1389,7 @@ class _AppHandler(object):
         """
         extensions = dict()
         location = 'app' if dname == self.app_dir else 'sys'
-        for target in glob.glob(pjoin(dname, 'extensions', '*.tgz')):
+        for target in glob(pjoin(dname, 'extensions', '*.tgz')):
             data = read_package(target)
             deps = data.get('dependencies', dict())
             name = data['name']
@@ -1342,7 +1425,14 @@ class _AppHandler(object):
         """
         compat = dict()
         core_data = self.info['core_data']
+        seen = set()
+        for (name, data) in self.info['federated_extensions'].items():
+            deps = data['dependencies']
+            compat[name] = _validate_compatibility(name, deps, core_data)
+            seen.add(name)
         for (name, data) in self.info['extensions'].items():
+            if name in seen:
+                continue
             deps = data['dependencies']
             compat[name] = _validate_compatibility(name, deps, core_data)
         return compat
@@ -1363,7 +1453,7 @@ class _AppHandler(object):
         if not osp.exists(dname):
             return info
 
-        for path in glob.glob(pjoin(dname, '*.tgz')):
+        for path in glob(pjoin(dname, '*.tgz')):
             path = osp.abspath(path)
             data = read_package(path)
             name = data['name']
@@ -1399,6 +1489,7 @@ class _AppHandler(object):
     def _list_extensions(self, info, ext_type):
         """List the extensions of a given type.
         """
+        self._ensure_disabled_info()
         logger = self.logger
         names = info['%s_extensions' % ext_type]
         if not names:
@@ -1410,6 +1501,8 @@ class _AppHandler(object):
 
         logger.info('   %s dir: %s' % (ext_type, dname))
         for name in sorted(names):
+            if name in info['federated_extensions']:
+                continue
             data = info['extensions'][name]
             version = data['version']
             errors = info['compat_errors'][name]
@@ -1436,6 +1529,57 @@ class _AppHandler(object):
         # Write all errors at end:
         _log_multiple_compat_errors(logger, error_accumulator)
 
+        # Write a blank line separator
+        logger.info('')
+
+    def _list_federated_extensions(self):
+        self._ensure_disabled_info()
+        info = self.info
+        logger = self.logger
+
+        error_accumulator = {}
+
+        ext_dirs = dict((p, False) for p in self.labextensions_path)
+        for value in info['federated_extensions'].values():
+            ext_dirs[value['ext_dir']] = True
+
+        for ext_dir, has_exts in ext_dirs.items():
+            if not has_exts:
+                continue
+            logger.info(ext_dir)
+            for name in info['federated_extensions']:
+                data = info['federated_extensions'][name]
+                if data['ext_dir'] != ext_dir:
+                    continue
+                version = data['version']
+                errors = info['compat_errors'][name]
+                extra = ''
+                if _is_disabled(name, info['disabled']):
+                    extra += ' %s' % RED_DISABLED
+                else:
+                    extra += ' %s' % GREEN_ENABLED
+                if errors:
+                    extra += ' %s' % RED_X
+                else:
+                    extra += ' %s' % GREEN_OK
+                if data['is_local']:
+                    extra += '*'
+
+                install = data.get('install')
+                if install:
+                    extra += ' (%s, %s)' % (
+                        install['packageManager'],
+                        install['packageName']
+                    )
+                logger.info('        %s v%s%s' % (name, version, extra))
+                if errors:
+                    error_accumulator[name] = (version, errors)
+            # Add a spacer line after
+            logger.info('')
+
+        # Write all errors at end:
+        _log_multiple_compat_errors(logger, error_accumulator)
+
     def _read_build_config(self):
         """Get the build config data for the app dir.
         """
@@ -1454,24 +1598,6 @@ class _AppHandler(object):
         with open(target, 'w') as fid:
             json.dump(config, fid, indent=4)
 
-    def _read_page_config(self):
-        """Get the page config data for the app dir.
-        """
-        target = pjoin(self.app_dir, 'settings', 'page_config.json')
-        if not osp.exists(target):
-            return {}
-        else:
-            with open(target) as fid:
-                return json.load(fid)
-
-    def _write_page_config(self, config):
-        """Write the build config to the app dir.
-        """
-        self._ensure_app_dirs()
-        target = pjoin(self.app_dir, 'settings', 'page_config.json')
-        with open(target, 'w') as fid:
-            json.dump(config, fid, indent=4)
-
     def _get_local_data(self, source):
         """Get the local data for extensions or linked packages.
         """
@@ -1578,7 +1704,7 @@ class _AppHandler(object):
             msg = '"%s" is not a valid npm package'
             raise ValueError(msg % source)
 
-        path = glob.glob(pjoin(tempdir, '*.tgz'))[0]
+        path = glob(pjoin(tempdir, '*.tgz'))[0]
         info['data'] = read_package(path)
         if is_dir:
             info['sha'] = sha = _tarsum(path)
